///////////////////////////////////////////////////////////////////////////////
// File		:	GameExporterPlugin.cpp
// Creator	:	Ashley Bennett
// Date		:	Saturday 10 September 2005
// Purpose	:	Plugin command class. Overloads the necessary functions to
// make a Maya plugin (MLL) work. Exports a Maya scene as a MRF file.
///////////////////////////////////////////////////////////////////////////////

//---INCLUDES---

#include <fstream>
using namespace std;

#include <maya/MSyntax.h>
#include <maya/MArgDatabase.h>
#include <maya/MGlobal.h>
#include <maya/MItDag.h>
#include <maya/MDagPath.h>
#include <maya/MFnMesh.h>
#include <maya/MPointArray.h>
#include <maya/MFloatArray.h>
#include <maya/MStringArray.h>
#include <maya/MItMeshPolygon.h>

#include "GameExporterPlugin.h"


//---PROTOTYPES---

//for texture function
void getTextureNamesFromMesh( MFnMesh &mesh, MStringArray &names, unsigned &count );


//---DEFINES---

//plugin flags
const char *fileFlag = "-f", *fileLongFlag = "-filename";
const char *typeFlag = "-t", *typeLongFlag = "-filetype";
const char *helpFlag = "-h", *helpLongFlag = "-help";

struct Vertex
{
	MPoint pos;
	MFloatVector normal;
	float u, v;
	bool visible;
};
struct Face
{
	//indexes to array positions holding the verticies making up a face
	unsigned int a, b, c;
};

//The kinds of files we can export
enum FILETYPES { FT_XML = 0, FT_BIN, NUMFILETYPES };


//---GameExporterPlugin CLASS FUNCTIONS---

///////////////////////////////////////////////////////////////////////////////
// creator()
//
// Creates a new instance of the plugin.
//
// Input:   NONE
///////////////////////////////////////////////////////////////////////////////
void *GameExporterPlugin::creator( void )
{
	return new GameExporterPlugin;
}


///////////////////////////////////////////////////////////////////////////////
// newSyntax()
//
// Syntax funtion to setup the flags used by this plugin.
//
// Input:   NONE
///////////////////////////////////////////////////////////////////////////////
MSyntax GameExporterPlugin::newSyntax( void )
{
    MSyntax syntax;

    syntax.addFlag( fileFlag, fileLongFlag, MSyntax::kString );
	syntax.addFlag( typeFlag, typeLongFlag, MSyntax::kUnsigned );
    syntax.addFlag( helpFlag, helpLongFlag );
    
    return syntax;
}


///////////////////////////////////////////////////////////////////////////////
// doIt()
//
// The meat of the plugin. Exports a 3D model in the format passed in the
// argument list, (default BIN.)
//
// Input:   args - Argument list
///////////////////////////////////////////////////////////////////////////////
MStatus GameExporterPlugin::doIt( const MArgList &args )
{
	//checking for flags (such as which file to output to)
	MArgDatabase argData( syntax(), args );
    MString fileName;
	UINT fileType;

    // was '-file' flag passed in?
    // if so catch the argument along with it
    if ( argData.isFlagSet( fileFlag ) )
        argData.getFlagArgument( fileFlag, 0, fileName );
	else
		fileName.set("C:\\Output.mrf");

	//make sure the filename wasn't blank
	if (fileName == "")
		fileName.set("C:\\Output.mrf");


    // was '-type' flag passed in?
    // if so catch the argument along with it
    if ( argData.isFlagSet( typeFlag ) )
        argData.getFlagArgument( typeFlag, 0, fileType );
	else
		fileType = FT_BIN;

	//make sure the filetype wasn't invalid
	if (fileType >= NUMFILETYPES)
		fileType = FT_BIN;


	// somewhere to catch return status
	MStatus status;
	// create dag iterator
	MItDag dagIterator( MItDag::kDepthFirst, MFn::kMesh, &status );
	// somewhere to save the path to a mesh
	MDagPath dagPath;

	// iterate through all meshes
	for ( ; !dagIterator.isDone(); dagIterator.next() )
	{
		// get the dagpath of current mesh
		dagIterator.getPath( dagPath );
	}


	// create mesh function set for mesh located at 'dagPath'
	MFnMesh mesh( dagPath, &status );

	// get points from the mesh
	MPointArray  points( mesh.numVertices() );
	mesh.getPoints( points, MSpace::kWorld );

	// get UVs from the mesh
	MFloatArray  meshUs( mesh.numUVs() ), meshVs( mesh.numUVs() );
	mesh.getUVs( meshUs, meshVs );

	// get normals from the mesh
	MFloatVectorArray  meshNormals( mesh.numNormals() );
	mesh.getNormals( meshNormals, MSpace::kWorld );

	// get all the textures
	MStringArray texNames;
	unsigned int texCount;
	getTextureNamesFromMesh( mesh, texNames, texCount );

	// somewhere to catch return status
	MStatus s;
	// create polygon iterator
	MItMeshPolygon faceIterator( dagPath, MObject::kNullObj, &s );

	// sizes of our arrays
	unsigned int numFaces = faceIterator.count();
	unsigned int size = numFaces * 3;
	// somewhere to save the faces
	Face* faceList = new Face[numFaces];
	// somewhere to save the vertices
	Vertex* vertexList = new Vertex[size];
	// keep track of how many verts we stored
	unsigned int vertCounter = 0;

	// iterate through all polygons
	for ( ; !faceIterator.isDone(); faceIterator.next() )
	{
		Vertex vertex[3];
		// get positions of all three verts per polygon
		// we are storing these in 'vertex' structures of our own
		vertex[ 0 ].pos = points[ faceIterator.vertexIndex( 0 ) ];
		vertex[ 1 ].pos = points[ faceIterator.vertexIndex( 1 ) ];
		vertex[ 2 ].pos = points[ faceIterator.vertexIndex( 2 ) ];
		
		//get the normals
		vertex[ 0 ].normal = meshNormals[ faceIterator.normalIndex( 0 ) ];
		vertex[ 1 ].normal = meshNormals[ faceIterator.normalIndex( 1 ) ];
		vertex[ 2 ].normal = meshNormals[ faceIterator.normalIndex( 2 ) ];
		
		//get the U and V's
		int index;
		faceIterator.getUVIndex( 0, index );
		vertex[ 0 ].u = meshUs[ index ];
		vertex[ 0 ].v = meshVs[ index ];
		faceIterator.getUVIndex( 1, index );
		vertex[ 1 ].u = meshUs[ index ];
		vertex[ 1 ].v = meshVs[ index ];
		faceIterator.getUVIndex( 2, index );
		vertex[ 2 ].u = meshUs[ index ];
		vertex[ 2 ].v = meshVs[ index ];


		//only add unique verts to the vertlist :

		//for each of our new points
		for (int v = 0; v < 3; v++)
		{
			//assume it is a new point
			vertex[v].visible = true;

			//for each point in our big list so far
			for (unsigned int vl = 0; vl < vertCounter; vl++)
			{
				//if the point is in the list already
				if (vertex[v].normal.x == vertexList[vl].normal.x &&
					vertex[v].normal.y == vertexList[vl].normal.y &&
					vertex[v].normal.z == vertexList[vl].normal.z &&
					vertex[v].pos.x == vertexList[vl].pos.x &&
					vertex[v].pos.y == vertexList[vl].pos.y &&
					vertex[v].pos.z == vertexList[vl].pos.z &&
					vertex[v].pos.w == vertexList[vl].pos.w &&
					vertex[v].u == vertexList[vl].u &&
					vertex[v].v == vertexList[vl].v)
				{
					//set it as hidden
					vertex[v].visible = false;
					break;
				}
			}

			//if it was a new point, add it to the points list
			if (vertex[v].visible)
				vertexList[vertCounter++] = vertex[v];
		}


		//only add unique verts to the facelist :

		//assume the points can't be found
		int faceIndex = faceIterator.index();
		faceList[faceIndex].a = -1;
		faceList[faceIndex].b = -1;
		faceList[faceIndex].c = -1;

		//find the first vertex
			//for each point in our big list
			for (unsigned int vl = 0; vl < vertCounter; vl++)
			{
				//if this point is the current point in the list
				if (vertex[0].normal.x == vertexList[vl].normal.x &&
					vertex[0].normal.y == vertexList[vl].normal.y &&
					vertex[0].normal.z == vertexList[vl].normal.z &&
					vertex[0].pos.x == vertexList[vl].pos.x &&
					vertex[0].pos.y == vertexList[vl].pos.y &&
					vertex[0].pos.z == vertexList[vl].pos.z &&
					vertex[0].pos.w == vertexList[vl].pos.w &&
					vertex[0].u == vertexList[vl].u &&
					vertex[0].v == vertexList[vl].v)
				{
					//set the face vert to this index
					faceList[faceIndex].a = vl;
					//stop looking for this vert
					break;
				}
			}

		//find the second vertex
			//for each point in our big list
			for (unsigned int vl = 0; vl < vertCounter; vl++)
			{
				//if this point is the current point in the list
				if (vertex[1].normal.x == vertexList[vl].normal.x &&
					vertex[1].normal.y == vertexList[vl].normal.y &&
					vertex[1].normal.z == vertexList[vl].normal.z &&
					vertex[1].pos.x == vertexList[vl].pos.x &&
					vertex[1].pos.y == vertexList[vl].pos.y &&
					vertex[1].pos.z == vertexList[vl].pos.z &&
					vertex[1].pos.w == vertexList[vl].pos.w &&
					vertex[1].u == vertexList[vl].u &&
					vertex[1].v == vertexList[vl].v)
				{
					//set the face vert to this index
					faceList[faceIndex].b = vl;
					//stop looking for this vert
					break;
				}
			}

		//find the third vertex
			//for each point in our big list
			for (unsigned int vl = 0; vl < vertCounter; vl++)
			{
				//if this point is the current point in the list
				if (vertex[2].normal.x == vertexList[vl].normal.x &&
					vertex[2].normal.y == vertexList[vl].normal.y &&
					vertex[2].normal.z == vertexList[vl].normal.z &&
					vertex[2].pos.x == vertexList[vl].pos.x &&
					vertex[2].pos.y == vertexList[vl].pos.y &&
					vertex[2].pos.z == vertexList[vl].pos.z &&
					vertex[2].pos.w == vertexList[vl].pos.w &&
					vertex[2].u == vertexList[vl].u &&
					vertex[2].v == vertexList[vl].v)
				{
					//set the face vert to this index
					faceList[faceIndex].c = vl;
					//stop looking for this vert
					break;
				}
			}

	}


	//Output the data in the requested format

	if (fileType == FT_XML)
	{
		// create an output stream to save the file
		ofstream fout;
		fout.open(fileName.asChar(), ios_base::trunc);

		// write out the intro
		fout << "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n\n";
		fout << "<mesh>\n";
		fout << "\t<name>" << mesh.name() << "</name>\n";
		fout << "\t<path>" << dagPath.fullPathName() << "</path>\n";
		fout << "\t<numVerts>" << vertCounter << "</numVerts>\n";
		fout << "\t<numFaces>" << numFaces << "</numFaces>\n";
		fout << "\t<numTextures>" << texCount << "</numTextures>\n";

		//write out all the textures
		fout << "\n\t<textureList>\n";
		for (unsigned int i = 0; i < texCount; i++)
			fout << "\t\t<texture>" << texNames[i] << "</texture>\n";
		fout << "\t</textureList>\n";

		// write out the vertices
		fout << "\n\t<vertList>\n";
		for (unsigned int i = 0; i < vertCounter; i++)
		{
			fout << "\t\t<vertex>\n";
			fout << "\t\t\t<vertX>" << vertexList[i].pos.x << "</vertX>\n";
			fout << "\t\t\t<vertY>" << vertexList[i].pos.y << "</vertY>\n";
			fout << "\t\t\t<vertZ>" << vertexList[i].pos.z << "</vertZ>\n";
			fout << "\t\t\t<vertNX>" << vertexList[i].normal.x << "</vertNX>\n";
			fout << "\t\t\t<vertNY>" << vertexList[i].normal.y << "</vertNY>\n";
			fout << "\t\t\t<vertNZ>" << vertexList[i].normal.z << "</vertNZ>\n";
			fout << "\t\t\t<vertU>" << vertexList[i].u << "</vertU>\n";
			fout << "\t\t\t<vertV>" << vertexList[i].v << "</vertV>\n";
			fout << "\t\t</vertex>\n";
		}
		fout << "\t</vertList>\n";

		// write out the faces
		fout << "\n\t<faceList>\n";
		for (unsigned int i = 0; i < numFaces; i++)
		{
			fout << "\t\t<face>\n";
			fout << "\t\t\t<faceA>" << faceList[i].a << "</faceA>\n";
			fout << "\t\t\t<faceB>" << faceList[i].b << "</faceB>\n";
			fout << "\t\t\t<faceC>" << faceList[i].c << "</faceC>\n";
			fout << "\t\t</face>\n";
		}
		fout << "\t</faceList>\n";

		// write out the outro
		fout << "</mesh>\n";

		// close file
		fout.close();
	}
	else if (fileType == FT_BIN)
	{
		// create an output stream to save the file
		ofstream fout;
		fout.open(fileName.asChar(), ios_base::trunc | ios_base::binary);

		/*
		// write out the intro
		//fout.write( reinterpret_cast<char*>(&mesh.name().length()), sizeof(mesh.name().length()) );
		//fout.write( mesh.name().asChar(), mesh.name().length() );

		//fout.write( reinterpret_cast<char*>(&dagPath.fullPathName().length()), sizeof(dagPath.fullPathName().length()) );
		//fout.write( dagPath.fullPathName().asChar(), dagPath.fullPathName().length() );

		fout.write( reinterpret_cast<char*>(&vertCounter), sizeof(vertCounter) );
		fout.write( reinterpret_cast<char*>(&numFaces), sizeof(numFaces) );
		fout.write( reinterpret_cast<char*>(&texCount), sizeof(texCount) );

		//write out all the textures
		for (unsigned int i = 0; i < texCount; i++)
		{
			//fout.write( reinterpret_cast<char*>(&texNames[i].length()), sizeof(texNames[i].length()) );
			//fout.write( texNames[i].asChar(), texNames[i].length() );
		}

		// write out the vertices
		for (unsigned int i = 0; i < vertCounter; i++)
		{
			fout.write( reinterpret_cast<char*>(&vertexList[i].pos.x), sizeof(vertexList[i].pos.x) );
			fout.write( reinterpret_cast<char*>(&vertexList[i].pos.y), sizeof(vertexList[i].pos.y) );
			fout.write( reinterpret_cast<char*>(&vertexList[i].pos.z), sizeof(vertexList[i].pos.z) );

			fout.write( reinterpret_cast<char*>(&vertexList[i].normal.x), sizeof(vertexList[i].normal.x) );
			fout.write( reinterpret_cast<char*>(&vertexList[i].normal.y), sizeof(vertexList[i].normal.y) );
			fout.write( reinterpret_cast<char*>(&vertexList[i].normal.z), sizeof(vertexList[i].normal.z) );

			fout.write( reinterpret_cast<char*>(&vertexList[i].u), sizeof(vertexList[i].u) );
			fout.write( reinterpret_cast<char*>(&vertexList[i].v), sizeof(vertexList[i].v) );
		}

		// write out the faces
		for (unsigned int i = 0; i < numFaces; i++)
		{
			fout.write( reinterpret_cast<char*>(&faceList[i].a), sizeof(faceList[i].a) );
			fout.write( reinterpret_cast<char*>(&faceList[i].a), sizeof(faceList[i].b) );
			fout.write( reinterpret_cast<char*>(&faceList[i].a), sizeof(faceList[i].c) );
		}
		*/

		//Write out array sizes
		fout.write( reinterpret_cast<char*>(&vertCounter), sizeof(vertCounter) );
		fout.write( reinterpret_cast<char*>(&numFaces), sizeof(numFaces) );
		fout.write( reinterpret_cast<char*>(&texCount), sizeof(texCount) );

		//Make new arrays in the same format as Game Models
		float	*vdata	= new float[vertCounter * 3];
		float	*ndata	= new float[vertCounter * 3];
		float	*uvdata	= new float[vertCounter * 2];
		UINT	*fdata	= new UINT[numFaces * 3];

		//save the vertices, normals and UVs
		for (unsigned int i = 0; i < vertCounter; i++)
		{
			vdata[(i * 3)]		= float( vertexList[i].pos.x );
			vdata[(i * 3) + 1]	= float( vertexList[i].pos.y );
			vdata[(i * 3) + 2]	= float( vertexList[i].pos.z );

			ndata[(i * 3)]		= vertexList[i].normal.x;
			ndata[(i * 3) + 1]	= vertexList[i].normal.y;
			ndata[(i * 3) + 2]	= vertexList[i].normal.z;

			uvdata[(i * 2)]		= vertexList[i].u;
			uvdata[(i * 2) + 1]	= vertexList[i].v;
		}

		//save the faces
		for (unsigned int i = 0; i < numFaces; i++)
		{
			fdata[(i * 3)]		= faceList[i].a;
			fdata[(i * 3) + 1]	= faceList[i].b;
			fdata[(i * 3) + 2]	= faceList[i].c;
		}

		//write out the arrays
		fout.write( reinterpret_cast<char*>(vdata),  vertCounter * 3 * sizeof(float) );
		fout.write( reinterpret_cast<char*>(ndata),  vertCounter * 3 * sizeof(float) );
		fout.write( reinterpret_cast<char*>(uvdata), vertCounter * 2 * sizeof(float) );
		fout.write( reinterpret_cast<char*>(fdata),  numFaces    * 3 * sizeof(UINT) );

		// close file
		fout.close();

		//cleanup
		delete [] vdata;
		delete [] ndata;
		delete [] uvdata;
		delete [] fdata;
	}

	//cleanup
	delete [] vertexList;
	delete [] faceList;

	return MS::kSuccess;
}
