///////////////////////////////////////////////////////////////////////////////
// File		:	TextureFunction.cpp
// Creator	:	Joe Del Rocco
// Date		:	Saturday 10 September 2005
// Purpose	:	Gets textures names from Maya meshes.
///////////////////////////////////////////////////////////////////////////////

#include <maya/MDagPath.h>
#include <maya/MFnMesh.h>
#include <maya/MStringArray.h>
#include <maya/MFnSet.h>
#include <maya/MItDependencyGraph.h>
#include <maya/MItMeshPolygon.h>
#include <maya/MPlug.h>
#include <maya/MPlugArray.h>

///////////////////////////////////////////////////////////////////////////////
// getTextureNamesFromMesh()
//
// Gets the texture names from all the sets of a mesh.
//
// Input:   mesh  - mesh to get from
//          names - returned textures
//          count - returned number of textures
///////////////////////////////////////////////////////////////////////////////
void getTextureNamesFromMesh( MFnMesh &mesh, MStringArray &names, unsigned &count )
{
    // local vars
    MStatus         status;
    MObjectArray    meshSets;
	MObjectArray    meshComps;
    unsigned        meshSetCount = 0;

    // get sets & components from mesh
    mesh.getConnectedSetsAndMembers( 0, meshSets, meshComps, true );
    meshSetCount = meshSets.length();
	if ( meshSetCount > 1 ) meshSetCount--;

    // init in-vars
    names.clear();
    count = 0;

    // get info from sets
	for ( unsigned i = 0; i < meshSetCount; i++ )
    {
		MObject set  = meshSets[i];
		MObject comp = meshComps[i];
		MFnSet  fnSet( set );

        // make sure we have a polygon set
        MItMeshPolygon meshSetIterator( mesh.dagPath(), comp, &status );
        if ( !status ) continue;

		//
        MFnDependencyNode   fnNode( set ); 
        MPlugArray          connectedPlugs;            

        // get shader plug
        MPlug shaderPlug = fnNode.findPlug( "surfaceShader", &status );
        //_MAYA_ASSERT( status, MString( "Failed to get shader plug for: " + mesh.partialPathName() ) );
        if ( !status ) continue;
        if ( shaderPlug.isNull() ) continue;

        // get shader node
        shaderPlug.connectedTo( connectedPlugs, true, false, &status );
        //_MAYA_ASSERT( status, MString( "Failed to get shader node for: " + mesh.partialPathName() ) );
        if ( !status ) continue;
        if ( connectedPlugs.length() != 1 ) continue;

        // get color plug
        MPlug colorPlug = MFnDependencyNode( connectedPlugs[0].node() ).findPlug("color", &status);
        //_MAYA_ASSERT( status, MString( "Failed to get color plug for: " + mesh.partialPathName() ) );
        if ( !status ) continue;
        
		// get iterator for traversing our color plug
		MItDependencyGraph itDG( colorPlug, MFn::kFileTexture, MItDependencyGraph::kUpstream, MItDependencyGraph::kBreadthFirst, MItDependencyGraph::kNodeLevel, &status );
        if ( !status ) continue;

		// disable automatic pruning so that we can locate a specific plug 
		itDG.disablePruningOnFilter();
        if ( itDG.isDone() ) continue;

        // get file plug
        MObject textureNode = itDG.thisNode();
		MPlug   filePlug    = MFnDependencyNode( textureNode ).findPlug( "fileTextureName" );
        MString textureName( "" );

        // log texture name
        if ( filePlug.getValue( textureName ) && textureName != "" )
        {
            names.append( textureName );            
            count++;
        }		
    }
}
